@@ -14,7 +14,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  *
  */
 
@@ -298,7 +298,7 @@ static void copy_context_hunks (char **line, size_t *linelen, ssize_t *got,
 }
 
 static void convert_unified_hunks_to_context (char **line, size_t *linelen,
-					      unsigned long *linenum)
+					      ssize_t *got, unsigned long *linenum)
 {
 	unsigned long orig_offset, orig_count = 0, new_offset, new_count = 0;
 	char **orig_line = NULL, **new_line = NULL;
@@ -312,7 +312,7 @@ static void convert_unified_hunks_to_context (char **line, size_t *linelen,
 	if (feof (stdin))
 		goto eof;
 
-	if (getline (line, linelen, stdin) == -1)
+	if ((*got = getline (line, linelen, stdin)) == -1)
 		goto eof;
 	++*linenum;
 
@@ -345,7 +345,7 @@ static void convert_unified_hunks_to_context (char **line, size_t *linelen,
 		       (new_linenum < new_count) || newline) {
 			int get_out = 0;
 
-			if (getline (line, linelen, stdin) == -1)
+			if ((*got = getline (line, linelen, stdin)) == -1)
 				/* Should write out everything to date? */
 				break;
 			++*linenum;
@@ -574,7 +574,7 @@ static void do_convert_to_context (void)
 			fwrite (line + 4, (size_t) got - 4, 1, stdout);
 			puts ("***************");
 			convert_unified_hunks_to_context (&line, &linelen,
-							  &linenum);
+							  &got, &linenum);
 		}
 	}
 
@@ -624,7 +624,7 @@ static void copy_unified_hunks (char **line, size_t *linelen, ssize_t *got,
 }
 
 static void convert_context_hunks_to_unified (char **line, size_t *linelen,
-					      unsigned long *linenum)
+					      ssize_t *got, unsigned long *linenum)
 {
 	int happy = 1;
 
@@ -638,7 +638,6 @@ static void convert_context_hunks_to_unified (char **line, size_t *linelen,
 		size_t *linelengths[2];
 		size_t n_lines[2];
 		size_t at[2];
-		ssize_t got;
 
 		n_lines[0] = n_lines[1] = 0;
 		linelengths[0] = linelengths[1] = NULL;
@@ -651,7 +650,8 @@ static void convert_context_hunks_to_unified (char **line, size_t *linelen,
 			if (feof (stdin))
 				goto eof;
 
-			if (getline (line, linelen, stdin) == -1)
+			*got = getline (line, linelen, stdin);
+			if (*got == -1)
 				goto eof;
 			++*linenum;
 
@@ -702,8 +702,8 @@ static void convert_context_hunks_to_unified (char **line, size_t *linelen,
 			memset (lines[i], 0, sizeof (char *) * line_count[i]);
 
 			for (lnum = 0; lnum < line_count[i]; lnum++) {
-				got = getline (line, linelen, stdin);
-				if (got == -1)
+				*got = getline (line, linelen, stdin);
+				if (*got == -1)
 					goto eof;
 				++*linenum;
 
@@ -717,10 +717,10 @@ static void convert_context_hunks_to_unified (char **line, size_t *linelen,
 					}
 				}
 
-				lines[i][lnum] = xmalloc ((size_t) got + 1);
-				memcpy (lines[i][lnum], *line, (size_t) got);
-				lines[i][lnum][got] = '\0';
-				linelengths[i][lnum] = (size_t) got;
+				lines[i][lnum] = xmalloc ((size_t) *got + 1);
+				memcpy (lines[i][lnum], *line, (size_t) *got);
+				lines[i][lnum][*got] = '\0';
+				linelengths[i][lnum] = (size_t) *got;
 				if (**line == ' ')
 					unchanged++;
 			}
@@ -741,6 +741,7 @@ static void convert_context_hunks_to_unified (char **line, size_t *linelen,
 		while (at[0] < n_lines[0] || at[1] < n_lines[1]) {
 			char *l[2] = { NULL, NULL };
 			size_t llen[2];
+			llen[0] = llen[1] = 0;
 			if (lines[0] && at[0] < n_lines[0]) {
 				l[0] = lines[0][at[0]];
 				llen[0] = linelengths[0][at[0]];
@@ -883,7 +884,7 @@ static void do_convert_to_unified (void)
 			printf ("+++ ");
 			fwrite (line + 4, (size_t) got - 4, 1, stdout);
 			convert_context_hunks_to_unified (&line, &linelen,
-							  &linenum);
+							  &got, &linenum);
 		} else {
 			fwrite (line, (size_t) got, 1, stdout);
 			got = getline (&line, &linelen, stdin);
@@ -1091,7 +1092,7 @@ read_timestamp (const char *timestamp, struct tm *result, long *zone)
 char *
 filename_from_header (const char *header)
 {
-	int first_space = strcspn (header, " \t\n");
+	int first_space = strcspn (header, " \t\n\r");
 	int h = first_space;
 	while (header[h] == ' ') {
 		int i;
@@ -1101,10 +1102,10 @@ filename_from_header (const char *header)
 		if (!read_timestamp (header + h + i, NULL, NULL))
 			break;
 		h += i + 1;
-		h += strcspn (header + h, " \t\n");
+		h += strcspn (header + h, " \t\n\r");
 	}
 
-	if (header[h] == '\n' && h > first_space)
+	if ((header[h] == '\n' || header[h] == '\r') && h > first_space)
 		/* If we couldn't see a date we recognized, but did see
 		   at least one space, split at the first. */
 		h = first_space;
