@@ -16,7 +16,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  *
  */
 
@@ -34,11 +34,16 @@
 #include <fnmatch.h>
 #include <getopt.h>
 #include <locale.h>
-#include <regex.h>
+#ifdef HAVE_PCRE2POSIX_H
+# include <pcre2posix.h>
+#else
+# include <regex.h>
+#endif
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <time.h>
+#include <ctype.h>
 
 #include "util.h"
 #include "diff.h"
@@ -52,7 +57,9 @@ struct range {
 enum line_numbering {
 	None = 0,
 	Before,
-	After
+	After,
+	OriginalBefore,
+	OriginalAfter
 };
 
 enum {
@@ -61,6 +68,13 @@ enum {
 	output_file
 } output_matching = output_none;
 
+enum {
+	only_match_all = 0,
+	only_match_rem,
+	only_match_add,
+	only_match_mod  /* bitwise rem | add */
+} only_matching = only_match_all;
+
 static struct patlist *pat_include = NULL;
 static struct patlist *pat_exclude = NULL;
 static struct range *hunks = NULL;
@@ -155,27 +169,64 @@ static int file_exists (const char *name, const char *timestamp)
 static int output_header_line (const char *line)
 {
 	char *fn;
-	int h = strcspn (line + 4, "\t\n");
 
-	fwrite (line, 1, 4, stdout);
+	if (strncmp (line, "diff", 4) == 0 && isspace (line[4])) {
+		size_t		args = 0;
+		const char	*end = line + 5, *begin = end, *ws = end;
+		printf ("%.5s", line);
+		while (*end != 0) {
+			if (isspace (*begin))
+				begin = end;
+			if (isspace (*end)) {
+				if (*begin == '-') {
+					if (isspace (begin[1]))
+						++args;
+					printf ("%.*s", (int)(end - ws), ws);
+				} else {
+					printf ("%.*s", (int)(begin - ws), ws);
+					if (args == 0 && old_prefix_to_add)
+						fputs (old_prefix_to_add,
+						       stdout);
+					if (args == 1 && new_prefix_to_add)
+						fputs (new_prefix_to_add,
+						       stdout);
+					++args;
+					fn = xstrndup (begin, end - begin);
+					fputs (stripped (fn,
+							 strip_components),
+					       stdout);
+					free (fn);
+				}
+				ws = begin = end;
+			}
+			++end;
+		}
+		printf ("%.*s", (int)(end - ws), ws);
+	} else if (strncmp (line, "---", 3) == 0 ||
+		   strncmp (line, "+++", 3) == 0) {
 
-	if (prefix_to_add)
-		fputs (prefix_to_add, stdout);
-	else {
-		if (old_prefix_to_add && strncmp (line, "---", 3) == 0)
-			fputs (old_prefix_to_add, stdout);
-		if (new_prefix_to_add && strncmp (line, "+++", 3) == 0)
-			fputs (new_prefix_to_add, stdout);
-	}
+		int h = strcspn (line + 4, "\t\n");
+		fwrite (line, 1, 4, stdout);
 
-	fn = xstrndup (line + 4, h);
-	fputs (stripped (fn, strip_components), stdout);
-	if (removing_timestamp)
-		putchar ('\n');
-	else
-		fputs (line + 4 + h, stdout);
+		if (prefix_to_add)
+			fputs (prefix_to_add, stdout);
+		else {
+			if (old_prefix_to_add && strncmp (line, "---", 3) == 0)
+				fputs (old_prefix_to_add, stdout);
+			if (new_prefix_to_add && strncmp (line, "+++", 3) == 0)
+				fputs (new_prefix_to_add, stdout);
+		}
 
-	free (fn);
+		fn = xstrndup (line + 4, h);
+		fputs (stripped (fn, strip_components), stdout);
+		if (removing_timestamp)
+			putchar ('\n');
+		else
+			fputs (line + 4 + h, stdout);
+
+		free (fn);
+	} else
+		fputs (line, stdout);
 	return 0;
 }
 
@@ -184,7 +235,7 @@ file_matches (void)
 {
 	int f = 0;
 	struct range *r;
-	
+
 	// See if the file range list includes this file.  -1UL is a
 	// wildcard.
 	for (r = files; r; r = r->next)
@@ -394,9 +445,9 @@ do_unified (FILE *f, char **header, unsigned int num_headers,
                                         unsigned int i;
                                         for (i = 0; i < num_headers - 2; i++)
                                                 output_header_line (header[i]);
-					if (number_lines != After)
+					if (number_lines != After && number_lines != OriginalAfter)
 						output_header_line (header[num_headers - 2]);
-					if (number_lines != Before)
+					if (number_lines != Before && number_lines != OriginalBefore)
 						output_header_line (header[num_headers - 1]);
 					header_displayed = 1;
 				}
@@ -423,8 +474,13 @@ do_unified (FILE *f, char **header, unsigned int num_headers,
 							 " Hunk #%lu, %s",
 							 hunknum, bestname);
 
-					fputs (clean_comments ? "\n" : trailing,
-					       output_to);
+					if (clean_comments)
+						fputs ("\n", output_to);
+					else {
+						int idx = trailing - *line;
+						fwrite (trailing, (size_t) got-idx,
+							1, output_to);
+					}
 					break;
 				case Before:
 					// Note the initial line number
@@ -443,6 +499,22 @@ do_unified (FILE *f, char **header, unsigned int num_headers,
 					     hunknum > 1))
 						fputs ("...\n", output_to);
 					break;
+				case OriginalBefore:
+					// Note the initial line number (original, no munge)
+					track_linenum = orig_offset;
+					if (!first_hunk ||
+					    (output_matching == output_file &&
+					     hunknum > 1))
+						fputs ("...\n", output_to);
+					break;
+				case OriginalAfter:
+					// Note the initial line number (original, no munge)
+					track_linenum = new_offset;
+					if (!first_hunk ||
+					    (output_matching == output_file &&
+					     hunknum > 1))
+						fputs ("...\n", output_to);
+					break;
 				}
 			} else if (mode == mode_filter)
 				// We are missing this hunk out, but
@@ -461,6 +533,10 @@ do_unified (FILE *f, char **header, unsigned int num_headers,
 		}
 
 		if (hunk_match && mode == mode_grep &&
+		    (only_matching == only_match_all
+		     || (**line == '-' && only_matching & only_match_rem)
+		    || (**line == '+' && only_matching & only_match_add)
+		    ) &&
 		    !regexecs (regex, num_regex, *line + 1, 0, NULL, 0)) {
 			if (output_matching == output_none) {
 				if (!displayed_filename) {
@@ -484,9 +560,9 @@ do_unified (FILE *f, char **header, unsigned int num_headers,
                                                 unsigned int i;
                                                 for (i = 0; i < num_headers - 2; i++)
                                                         output_header_line (header[i]);
-                                                if (number_lines != After)
+                                                if (number_lines != After && number_lines != OriginalAfter)
                                                         output_header_line (header[num_headers - 2]);
-                                                if (number_lines != Before)
+                                                if (number_lines != Before && number_lines != OriginalBefore)
                                                         output_header_line (header[num_headers - 1]);
 						header_displayed = 1;
                                         }
@@ -515,10 +591,18 @@ do_unified (FILE *f, char **header, unsigned int num_headers,
 				// Just display each line.
 				fwrite (*line, (size_t) got, 1, output_to);
 			else if ((number_lines == Before && **line != '+') ||
-				 (number_lines == After && **line != '-'))
+				 (number_lines == After && **line != '-') ||
+				 (number_lines == OriginalBefore && **line != '+') ||
+				 (number_lines == OriginalAfter && **line != '-')) {
 				// Numbered line.
+                                const char *rest = *line;
+                                if (rest[0] != '\n')
+                                        // Handle whitespace damage
+                                        rest++;
+
 				fprintf (output_to, "%lu\t:%s",
-					 track_linenum++, 1 + *line);
+					 track_linenum++, rest);
+                        }
 		}
 	}
 
@@ -695,9 +779,9 @@ do_context (FILE *f, char **header, unsigned int num_headers,
                                 unsigned int i;
                                 for (i = 0; i < num_headers - 2; i++)
                                         output_header_line (header[i]);
-				if (number_lines != After)
+				if (number_lines != After && number_lines != OriginalAfter)
 					output_header_line (header[num_headers - 2]);
-				if (number_lines != Before)
+				if (number_lines != Before && number_lines != OriginalBefore)
 					output_header_line (header[num_headers - 1]);
 				header_displayed = 1;
 			}
@@ -756,12 +840,33 @@ do_context (FILE *f, char **header, unsigned int num_headers,
 				     hunknum > 1))
 					fputs ("...\n", output_to);
 				break;
+
+			case OriginalBefore:
+				if (i != 0)
+					break;
+
+				// Note the initial line number (original, no munge).
+				track_linenum = line_start;
+				if (!first_hunk ||
+				    (output_matching == output_file &&
+				     hunknum > 1))
+					fputs ("...\n", output_to);
+				break;
+
+			case OriginalAfter:
+				if (i != 1)
+					break;
+
+				// Note the initial line number (original, no munge).
+				track_linenum = line_start;
+				if (!first_hunk ||
+				    (output_matching == output_file &&
+				     hunknum > 1))
+					fputs ("...\n", output_to);
+				break;
 			}
 		}
 
-		if (i && line_count == unchanged)
-			break;
-
 		got = getline (line, linelen, f);
 		if (got == -1) {
 			ret = EOF;
@@ -770,9 +875,16 @@ do_context (FILE *f, char **header, unsigned int num_headers,
 
 		++*linenum;
 
+		if (i && line_count == unchanged)
+			break;
+
 		while ((line_count == 0 && **line == '\\') ||
 		       line_count--) {
 			if (hunk_match && mode == mode_grep &&
+			    (only_matching == only_match_all
+			     || (**line != ' ' && i == 0 && only_matching & only_match_rem)
+			     || (**line != ' ' && i == 1 && only_matching & only_match_add)
+			    ) &&
 			    !regexecs (regex, num_regex, *line + 2,
 				       0, NULL, 0)) {
 				if (output_matching == output_none) {
@@ -798,9 +910,9 @@ do_context (FILE *f, char **header, unsigned int num_headers,
                                                 unsigned int i;
                                                 for (i = 0; i < num_headers - 2; i++)
                                                         output_header_line (header[i]);
-						if (number_lines != After)
+						if (number_lines != After && number_lines != OriginalAfter)
 							output_header_line (header[num_headers - 2]);
-						if (number_lines != Before)
+						if (number_lines != Before && number_lines != OriginalBefore)
 							output_header_line (header[num_headers - 1]);
 						header_displayed = 1;
 					}
@@ -859,7 +971,9 @@ do_context (FILE *f, char **header, unsigned int num_headers,
 					fwrite (*line, (size_t) got,
 						1, output_to);
 				else if ((number_lines == Before && !i) ||
-					 (number_lines == After && i)) {
+					 (number_lines == After && i) ||
+					 (number_lines == OriginalBefore && !i) ||
+					 (number_lines == OriginalAfter && i)) {
 					fprintf (output_to, "%lu\t:",
 						 track_linenum++);
 					fwrite (2 + *line, (size_t) got - 2,
@@ -910,12 +1024,12 @@ out:
 	return ret;
 }
 
-#define MAX_HEADERS 5
+#define MAX_HEADERS 6
 static int filterdiff (FILE *f, const char *patchname)
 {
 	static unsigned long linenum = 1;
 	char *names[2];
-	char *header[MAX_HEADERS] = { NULL, NULL };
+	char *header[MAX_HEADERS + 2] = { NULL, NULL };
         unsigned int num_headers = 0;
 	char *line = NULL;
 	size_t linelen = 0;
@@ -991,7 +1105,7 @@ static int filterdiff (FILE *f, const char *patchname)
                                 else if (strncmp (line, "old mode ", 9) &&
                                     strncmp (line, "new mode ", 9) &&
                                     strncmp (line, "deleted file mode ", 18) &&
-                                    strncmp (line, "new file mode ", 15) &&
+                                    strncmp (line, "new file mode ", 14) &&
                                     strncmp (line, "copy from ", 10) &&
                                     strncmp (line, "copy to ", 8) &&
                                     strncmp (line, "rename from ", 12) &&
@@ -1005,8 +1119,10 @@ static int filterdiff (FILE *f, const char *patchname)
                                         break;
 
                                 /* Drop excess header lines */
-                                if (num_headers < MAX_HEADERS - 2)
-                                        header[num_headers++] = xstrdup (line);
+                                if (num_headers > MAX_HEADERS )
+                                        free (header[--num_headers]);
+
+                                header[num_headers++] = xstrdup (line);
 
                                 if (is_context != -1)
                                         break;
@@ -1145,19 +1261,24 @@ const char * syntax_str =
 "  --lines=L include only hunks with (original) lines in range L, if range begins with x show all excluding range L\n"
 "  -F F, --files=F\n"
 "            include only files in range F, if range begins with x show all excluding range F\n"
-"  --annotate (filterdiff, grepdiff)\n"
-"            annotate each hunk with the filename and hunk number (filterdiff, grepdiff)\n"
-"  --as-numbered-lines=before|after (filterdiff, grepdiff)\n"
-"            display lines as they would look before, or after, the (filterdiff, grepdiff)\n"
-"            patch is applied (filterdiff, grepdiff)\n"
-"  --format=context|unified (filterdiff, grepdiff)\n"
-"            set output format (filterdiff, grepdiff)\n"
+"  --annotate (filterdiff, patchview, grepdiff)\n"
+"            annotate each hunk with the filename and hunk number (filterdiff, patchview, grepdiff)\n"
+"  --as-numbered-lines=before|after|original-before|original-after (filterdiff, patchview, grepdiff)\n"
+"            display lines as they would look before, or after, the patch is applied;\n"
+"            or with original line numbers from the diff (original-before/original-after)\n"
+"  --format=context|unified (filterdiff, patchview, grepdiff)\n"
+"            set output format (filterdiff, patchview, grepdiff)\n"
 "  --output-matching=hunk|file (grepdiff)\n"
 "            show matching hunks or file-level diffs (grepdiff)\n"
-"  --remove-timestamps (filterdiff, grepdiff)\n"
-"            don't show timestamps from output (filterdiff, grepdiff)\n"
+"  --only-match=rem|removals|add|additions|mod|modifications|all (grepdiff)\n"
+"            regex will only match removals, additions, modifications or (grepdiff)\n"
+"            the whole hunk (grepdiff)\n"
+"  --remove-timestamps (filterdiff, patchview, grepdiff)\n"
+"            don't show timestamps from output (filterdiff, patchview, grepdiff)\n"
 "  --clean (filterdiff)\n"
 "            remove all comments (non-diff lines) from output (filterdiff)\n"
+"  --in-place (filterdiff)\n"
+"            write output to the original input files (filterdiff)\n"
 "  -z, --decompress\n"
 "            decompress .gz and .bz2 files\n"
 "  -n, --line-number (lsdiff, grepdiff)\n"
@@ -1182,14 +1303,18 @@ const char * syntax_str =
 "  -v, --verbose\n"
 "            verbose output -- use more than once for extra verbosity\n"
 "  -E, --extended-regexp (grepdiff)\n"
+#ifdef HAVE_PCRE2POSIX_H
+"            this option has no effect as PCRE regexes are used by default (grepdiff)\n"
+#else
 "            use extended regexps, like egrep (grepdiff)\n"
+#endif
 "  -E, --empty-files-as-absent (lsdiff)\n"
 "            treat empty files as absent (lsdiff)\n"
 "  -f FILE, --file=FILE (grepdiff)\n"
 "            read regular expressions from FILE (grepdiff)\n"
-"  --filter  run as 'filterdiff' (grepdiff, lsdiff)\n"
-"  --list    run as 'lsdiff' (filterdiff, grepdiff)\n"
-"  --grep    run as 'grepdiff' (filterdiff, lsdiff)\n"
+"  --filter  run as 'filterdiff' (grepdiff, patchview, lsdiff)\n"
+"  --list    run as 'lsdiff' (filterdiff, patchview, grepdiff)\n"
+"  --grep    run as 'grepdiff' (filterdiff, patchview, lsdiff)\n"
 ;
 
 NORETURN
@@ -1295,6 +1420,13 @@ static void set_filter (void)
 	mode = mode_filter;
 }
 
+static void set_view (void)
+{
+	/* This is patchview. */
+	set_progname ("patchview");
+	mode = mode_filter;
+}
+
 static void set_grep (void)
 {
 	/* This is grepdiff. */
@@ -1304,7 +1436,7 @@ static void set_grep (void)
 
 static void determine_mode_from_name (const char *argv0)
 {
-	/* This is filterdiff, unless it is named 'lsdiff' or 'grepdiff'. */
+	/* This is filterdiff, unless it is named 'lsdiff', 'grepdiff' or 'patchview'. */
 	const char *p = strrchr (argv0, '/');
 	if (!p++)
 		p = argv0;
@@ -1312,6 +1444,9 @@ static void determine_mode_from_name (const char *argv0)
 		set_list ();
 	else if (strstr (p, "grepdiff"))
 		set_grep ();
+	else if (strstr (p, "patchview")) {
+		set_view ();
+	}
 	else
 		set_filter ();
 }
@@ -1355,7 +1490,7 @@ read_regex_file (const char *file)
 			char errstr[300];
 			regerror (err, &regex[num_regex - 1], errstr,
 				  sizeof (errstr));
-			error (EXIT_FAILURE, 0, errstr);
+			error (EXIT_FAILURE, 0, "%s", errstr);
 			exit (1);
 		}
 	}
@@ -1370,6 +1505,8 @@ int main (int argc, char *argv[])
 	FILE *f = stdin;
 	char format = '\0';
 	int regex_file_specified = 0;
+	int have_switches = 0;
+	int inplace_mode = 0;
 
 	setlocale (LC_TIME, "C");
 	determine_mode_from_name (argv[0]);
@@ -1392,6 +1529,7 @@ int main (int argc, char *argv[])
 			{"annotate", 0, 0, 1000 + 'a'},
 			{"format", 1, 0, 1000 + 'F'},
 			{"output-matching", 1, 0, 1000 + 'o'},
+			{"only-match", 1, 0, 1000 + 'm'},
 			{"remove-timestamps", 0, 0, 1000 + 'r'},
 			{"with-filename", 0, 0, 'H'},
 			{"no-filename", 0, 0, 'h'},
@@ -1410,6 +1548,7 @@ int main (int argc, char *argv[])
 			{"extended-regexp", 0, 0, 'E'},
 			{"empty-files-as-removed", 0, 0, 'E'},
 			{"file", 1, 0, 'f'},
+			{"in-place", 0, 0, 1000 + 'w'},
 			{0, 0, 0, 0}
 		};
 		char *end;
@@ -1417,7 +1556,8 @@ int main (int argc, char *argv[])
 				     long_options, NULL);
 		if (c == -1)
 			break;
-		
+
+		have_switches = 1;
 		switch (c) {
 		case 'g':
 			set_grep ();
@@ -1538,6 +1678,10 @@ int main (int argc, char *argv[])
 				number_lines = Before;
 			else if (!strcmp (optarg, "after"))
 				number_lines = After;
+			else if (!strcmp (optarg, "original-before"))
+				number_lines = OriginalBefore;
+			else if (!strcmp (optarg, "original-after"))
+				number_lines = OriginalAfter;
 			else syntax (1);
 			break;
 		case 1000 + 'a':
@@ -1559,16 +1703,37 @@ int main (int argc, char *argv[])
 				output_matching = output_file;
 			else syntax (1);
 			break;
+		case 1000 + 'm':
+			if (!strncmp (optarg, "all", 3))
+				only_matching = only_match_all;
+			else if (!strncmp (optarg, "rem", 3) ||
+			         !strncmp (optarg, "removals", 8))
+				only_matching = only_match_rem;
+			else if (!strncmp (optarg, "add", 3) ||
+			         !strncmp (optarg, "additions", 9))
+				only_matching = only_match_add;
+			else if (!strncmp (optarg, "mod", 3) ||
+			         !strncmp (optarg, "modifications", 13))
+				only_matching = only_match_mod;
+			else syntax (1);
+			break;
 		case 1000 + 'r':
 			removing_timestamp = 1;
 			break;
 		case 1000 + 'c':
 			clean_comments = 1;
 			break;
+		case 1000 + 'w':
+			inplace_mode = 1;
+			break;
 		default:
 			syntax(1);
 		}
 	}
+    if (have_switches == 0 && strcmp (progname, "patchview") == 0) {
+        mode = mode_list;
+        number_files = 1;
+    }
 
 	/* Preserve the old semantics of -p. */
 	if (mode != mode_filter && ignore_components && !strip_components &&
@@ -1584,6 +1749,10 @@ int main (int argc, char *argv[])
 		error (EXIT_FAILURE, 0, "--output-matching only applies to "
 		       "grep mode");
 
+	if (mode != mode_grep && only_matching != only_match_all)
+		error (EXIT_FAILURE, 0, "--only-match only applies to "
+		       "grep mode");
+
 	if (numbering &&
 	    !(mode == mode_list ||
 	      (mode == mode_grep && output_matching == output_none)))
@@ -1600,6 +1769,14 @@ int main (int argc, char *argv[])
 		error (EXIT_FAILURE, 0, "can't use --verbose and "
 		       "--clean options simultaneously");
 
+	if (inplace_mode && unzip)
+		error (EXIT_FAILURE, 0,
+		       "--in-place and --decompress are mutually exclusive");
+
+	if (inplace_mode && mode != mode_filter)
+		error (EXIT_FAILURE, 0,
+		       "--in-place only applies to filter mode");
+
 	if (mode == mode_grep && !regex_file_specified) {
 		int err;
 
@@ -1613,7 +1790,7 @@ int main (int argc, char *argv[])
 			char errstr[300];
 			regerror (err, &regex[num_regex - 1], errstr,
 				  sizeof (errstr));
-			error (EXIT_FAILURE, 0, errstr);
+			error (EXIT_FAILURE, 0, "%s", errstr);
 			exit (1);
 		}
 	}
@@ -1631,6 +1808,10 @@ int main (int argc, char *argv[])
 			print_patchnames = 0;
 	}
 
+	if (inplace_mode && optind == argc)
+		error (EXIT_FAILURE, 0,
+		       "--in-place cannot be used with standard input");
+
 	if (optind == argc) {
 		f = convert_format (stdin, format);
 		filterdiff (f, "(standard input)");
@@ -1644,7 +1825,28 @@ int main (int argc, char *argv[])
 			}
 
 			f = convert_format (f, format);
-			filterdiff (f, argv[i]);
+
+			if (inplace_mode) {
+				/* Redirect stdout to temporary file for in-place processing */
+				FILE *temp_output = xtmpfile();
+				FILE *old_stdout = stdout;
+				stdout = temp_output;
+
+				filterdiff (f, argv[i]);
+
+				/* Restore stdout */
+				stdout = old_stdout;
+
+				/* Write temp file contents back to original file */
+				if (write_file_inplace(argv[i], temp_output) != 0) {
+					error (EXIT_FAILURE, errno, "failed to write %s", argv[i]);
+				}
+
+				fclose (temp_output);
+			} else {
+				filterdiff (f, argv[i]);
+			}
+
 			fclose (f);
 		}
 	}
