@@ -1,6 +1,6 @@
 /*
  * rediff - fix offset and counts of a hand-edited diff
- * Copyright (C) 2001, 2002, 2004, 2009, 2011 Tim Waugh <twaugh@redhat.com>
+ * Copyright (C) 2001, 2002, 2004, 2009, 2011, 2019 Tim Waugh <twaugh@redhat.com>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -14,7 +14,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  *
  */
 
@@ -51,13 +51,14 @@
 
 struct file_info
 {
+	const char *diff_command;
 	const char *orig_file;
 	const char *new_file;
 	int info_written;
 	int info_pending;
 };
 
-struct hunk 
+struct hunk
 {
 	fpos_t filepos;
 	struct file_info *info;
@@ -121,7 +122,7 @@ static unsigned long copy_hunk (FILE *in, FILE *out,
 
 /* Copy hunk from in to out, adjusting offsets by line_offset. */
 static unsigned long adjust_offsets_and_copy (long *offset, FILE *in,
-					      FILE *out)
+					      FILE *out, struct hunk *hunk)
 {
 	char *line = NULL;
 	size_t linelen = 0;
@@ -136,6 +137,8 @@ static unsigned long adjust_offsets_and_copy (long *offset, FILE *in,
 	if (!strncmp (line, "--- ", 4)) {
 		/* This is the first hunk of a group.  Copy the
 		 * file info. */
+		if (hunk && hunk->info && hunk->info->diff_command)
+			fputs (hunk->info->diff_command, out);
 		fputs (line, out);
 		if (getline (&line, &linelen, in) == -1)
 			goto out;
@@ -228,21 +231,35 @@ static void copy_to (struct hunk *from, struct hunk *upto,
 {
 	if (!is_first && from && from->info && !from->info->info_written &&
 	    from->info->info_pending) {
+		if (from->info->diff_command)
+			fputs (from->info->diff_command, out);
 		fputs (from->info->orig_file, out);
 		fputs (from->info->new_file, out);
 		from->info->info_written = 1;
 	}
 
 	if (is_first && from) {
-		/* Copy leading non-diff text. */
+		/* Copy leading non-diff text, but skip diff command lines
+		 * as they will be output by the file info mechanism. */
 		fseek (in, 0, SEEK_SET);
-		copy_lines (in, out, from->line_in_diff - 1);
+		unsigned long lines_to_copy = from->line_in_diff - 1;
+		char *line = NULL;
+		size_t linelen = 0;
+		for (unsigned long i = 0; i < lines_to_copy; i++) {
+			if (getline (&line, &linelen, in) == -1)
+				break;
+			/* Skip diff command lines as they're handled by file info */
+			if (strncmp (line, "diff ", 5) != 0)
+				fputs (line, out);
+		}
+		if (line)
+			free (line);
 	}
 
 	for (; from && from != upto; from = from->next) {
 		unsigned long count;
 		fsetpos (in, &from->filepos);
-		count = adjust_offsets_and_copy (line_offset, in, out);
+		count = adjust_offsets_and_copy (line_offset, in, out, from);
 		copy_trailing (from, in, out, count);
 	}
 }
@@ -358,6 +375,8 @@ static long removed_hunk (const char *meta, FILE *modify, FILE *t,
 
 		/* Display a file name banner. */
 		if (hunk->info && !hunk->info->info_written) {
+			if (hunk->info->diff_command)
+				fputs (hunk->info->diff_command, t);
 			fputs (hunk->info->orig_file, t);
 			fputs (hunk->info->new_file, t);
 			hunk->info->info_written = 1;
@@ -511,7 +530,7 @@ static long show_modified_hunk (struct hunk **hunkp, long line_offset,
 
 	do {
 		/* Lines before the modification are unaltered. */
-		int trim = 0;
+                int trim = 0, skip_trim;
 		if (morig_offset < hunk->line_in_diff)
 			error (EXIT_FAILURE, errno, "Invalid changes made");
 
@@ -547,6 +566,7 @@ static long show_modified_hunk (struct hunk **hunkp, long line_offset,
 		}
 
 		while (morig_count || mnew_count) {
+                        skip_trim = 0;
 			if (getline (&line, &linelen, modify) == -1)
 				break;
 
@@ -564,7 +584,7 @@ static long show_modified_hunk (struct hunk **hunkp, long line_offset,
 				case '+':
 					this_offset--;
 					calc_new_count--;
-					trim = 0;
+					skip_trim = 1;
 					break;
 				case '-':
 					this_offset++;
@@ -582,7 +602,7 @@ static long show_modified_hunk (struct hunk **hunkp, long line_offset,
 					       line[0], line[1]);
 				}
 
-				if (trim) {
+				if (trim && !skip_trim) {
 					orig_offset++;
 					calc_new_offset++;
 				}
@@ -660,6 +680,8 @@ static long show_modified_hunk (struct hunk **hunkp, long line_offset,
 				FILE *write_to;
 				write_to = t_written_to ? t : out;
 				if (hunk->info && !hunk->info->info_written) {
+					if (hunk->info->diff_command)
+						fputs (hunk->info->diff_command, out);
 					fputs (hunk->info->orig_file, out);
 					fputs (hunk->info->new_file, out);
 					hunk->info->info_written = 1;
@@ -758,6 +780,8 @@ static long show_modified_hunk (struct hunk **hunkp, long line_offset,
 #endif /* DEBUG */
 
 	if (hunk->info && !hunk->info->info_written) {
+		if (hunk->info->diff_command)
+			fputs (hunk->info->diff_command, out);
 		fputs (hunk->info->orig_file, out);
 		fputs (hunk->info->new_file, out);
 		hunk->info->info_written = 1;
@@ -802,7 +826,8 @@ static int rediff (const char *original, const char *edited, FILE *out)
 	pid_t child;
 	FILE *o;
 	FILE *m;
-	FILE *t = NULL;
+	FILE *diff_pipe = NULL;
+	FILE *meta_hunks = NULL;
 	char *line = NULL;
 	size_t linelen = 0;
 	unsigned long linenum = 0;
@@ -816,6 +841,7 @@ static int rediff (const char *original, const char *edited, FILE *out)
 		unsigned long o_count, n_count;
 		struct hunk *newhunk;
 		fpos_t pos;
+		char *diff_cmd = NULL;
 
 		/* Search for start of hunk (or file info). */
 		do {
@@ -828,6 +854,13 @@ static int rediff (const char *original, const char *edited, FILE *out)
 				error (EXIT_FAILURE, errno,
 				       "Don't know how to handle context "
 				       "format yet.");
+
+			/* Capture potential diff command line */
+			if (!strncmp (line, "diff ", 5)) {
+				if (diff_cmd)
+					free (diff_cmd);
+				diff_cmd = xstrdup (line);
+			}
 		} while (strncmp (line, "@@ ", 3) &&
 			 strncmp (line, "--- ", 4));
 
@@ -845,6 +878,7 @@ static int rediff (const char *original, const char *edited, FILE *out)
 		if (!strncmp (line, "--- ", 4)) {
 			struct file_info *info = xmalloc (sizeof *info);
 			info->info_written = info->info_pending = 0;
+			info->diff_command = diff_cmd;
 			info->orig_file = xstrdup (line);
 			if (getline (&line, &linelen, o) == -1)
 				error (EXIT_FAILURE, errno,
@@ -855,7 +889,11 @@ static int rediff (const char *original, const char *edited, FILE *out)
 				error (EXIT_FAILURE, errno,
 				       "Premature end of file");
 			linenum += 2;
-		} else newhunk->info = NULL;
+		} else {
+			newhunk->info = NULL;
+			if (diff_cmd)
+				free (diff_cmd);
+		}
 
 		read_atatline (line, &newhunk->orig_offset,
 			       &newhunk->orig_count,
@@ -904,17 +942,17 @@ static int rediff (const char *original, const char *edited, FILE *out)
 	last->num_lines = linenum - last->line_in_diff + 1;
 
 	/* Run diff between original and edited. */
-	t = xpipe (DIFF, &child, "r", DIFF, "-U0",
-		   original, edited, NULL);
+	diff_pipe = xpipe (DIFF, &child, "r", (char **) (const char*[]) {DIFF, "-U0",
+		   original, edited, NULL });
 	m = xtmpfile ();
 	if (m) {
 		size_t buffer_size = 10000;
 		char *buffer = xmalloc (buffer_size);
-		while (!feof (t)) {
-			size_t got = fread (buffer, 1, buffer_size, t);
+		while (!feof (diff_pipe)) {
+			size_t got = fread (buffer, 1, buffer_size, diff_pipe);
 			fwrite (buffer, 1, got, m);
 		}
-		fclose (t);
+		fclose (diff_pipe);
 		waitpid (child, NULL, 0);
 		rewind (m);
 		free (buffer);
@@ -936,7 +974,7 @@ static int rediff (const char *original, const char *edited, FILE *out)
 			if (getline (&line, &linelen, m) == -1)
 				break;
 		}
-	
+
 		if (feof (m))
 			break;
 
@@ -977,7 +1015,7 @@ static int rediff (const char *original, const char *edited, FILE *out)
 			if (current_hunk) {
 				line_offset += show_modified_hunk
 					(&current_hunk, line_offset,
-					 t, o, out);
+					 meta_hunks, o, out);
 				current_hunk = current_hunk->next;
 			}
 
@@ -988,26 +1026,26 @@ static int rediff (const char *original, const char *edited, FILE *out)
 
 			/* This meta hunk is the first pertaining to
 			 * the hunk in the original. */
-			t = xtmpfile ();
+			meta_hunks = xtmpfile ();
 		}
 
 		current_hunk = which;
 
 		/* Append the meta hunk to a temporary file. */
-		fputs (line, t);
+		fputs (line, meta_hunks);
 		while (!feof (m)) {
 			if (getline (&line, &linelen, m) == -1)
 				break;
 			if (!strncmp (line, "@@ ", 3))
 				break;
-			fputs (line, t);
+			fputs (line, meta_hunks);
 		}
 	}
 
 	/* Now display the remaining hunks, adjusting offsets. */
 	if (current_hunk) {
 		line_offset += show_modified_hunk (&current_hunk, line_offset,
-						   t, o, out);
+						   meta_hunks, o, out);
 		current_hunk = current_hunk->next;
 		if (current_hunk)
 			copy_to (current_hunk, NULL, &line_offset, o, out, 0);
@@ -1021,8 +1059,8 @@ static int rediff (const char *original, const char *edited, FILE *out)
 
 	return 0;
 }
-static char * syntax_str = "usage: %s ORIGINAL EDITED\n"
-                           "       %s EDITED\n";
+static char * syntax_str = "usage: %s [--in-place] ORIGINAL EDITED\n"
+                           "       %s [--in-place] EDITED\n";
 
 NORETURN
 static void syntax (int err)
@@ -1035,18 +1073,20 @@ int main (int argc, char *argv[])
 {
 	/* name to use in error messages */
 	set_progname ("rediff");
-	
+	int inplace_mode = 0;
+
 	while (1) {
 		static struct option long_options[] = {
 	       		{"help", 0, 0, 'h'},
 			{"version", 0, 0, 'v'},
+			{"in-place", 0, 0, 1000 + 'w'},
 			{0, 0, 0, 0}
 		};
 		int c = getopt_long (argc, argv, "vh",
 				long_options, NULL);
 		if (c == -1)
 			break;
-		
+
 		switch (c) {
 		case 'v':
 			printf("rediff - patchutils version %s\n", VERSION);
@@ -1054,45 +1094,68 @@ int main (int argc, char *argv[])
 		case 'h':
 			syntax (0);
 			break;
+		case 1000 + 'w':
+			inplace_mode = 1;
+			break;
 		default:
 			syntax(1);
 		}
-				
+
 	}
-	
+
 	if (argc - optind < 1)
 		syntax (1);
 
 	if (argc - optind == 1) {
-		char *p = xmalloc (strlen (argv[0]) +
-				   strlen ("recountdiff") + 1);
-		char *f;
-		char **const new_argv = xmalloc (sizeof (char *) * (argc + 1));
-		memcpy (new_argv, argv, sizeof (char *) * (argc + 1));
-		new_argv[0] = p;
-		strcpy (p, argv[0]);
-		f = strrchr (p, '/');
-		if (!f)
-			f = p;
-		else f++;
-		strcpy (f, "recountdiff");
-		execvp (new_argv[0], new_argv);
-		p = xstrdup (new_argv[0]);
-		f = strstr (p, "src/");
-		if (f) {
-			while (*(f + 4)) {
-				*f = *(f + 4);
-				f++;
-			}
-			*f = '\0';
+		if (inplace_mode) {
+			/* For single argument with --in-place, we need to handle recountdiff differently */
+			error (EXIT_FAILURE, 0, "--in-place with single argument not yet implemented");
+		} else {
+			char *p = xmalloc (strlen (argv[0]) +
+					   strlen ("recountdiff") + 1);
+			char *f;
+			char **const new_argv = xmalloc (sizeof (char *) * (argc + 1));
+			memcpy (new_argv, argv, sizeof (char *) * (argc + 1));
 			new_argv[0] = p;
-			execv (new_argv[0], new_argv);
+			strcpy (p, argv[0]);
+			f = strrchr (p, '/');
+			if (!f)
+				f = p;
+			else f++;
+			strcpy (f, "recountdiff");
+			execvp (new_argv[0], new_argv);
+			p = xstrdup (new_argv[0]);
+			f = strstr (p, "src/");
+			if (f) {
+				while (*(f + 4)) {
+					*f = *(f + 4);
+					f++;
+				}
+				*f = '\0';
+				new_argv[0] = p;
+				execv (new_argv[0], new_argv);
+			}
+			error (EXIT_FAILURE, 0, "couldn't execute recountdiff");
 		}
-		error (EXIT_FAILURE, 0, "couldn't execute recountdiff");
 	}
 
 	if (access (argv[optind + 1], R_OK))
 		error (EXIT_FAILURE, errno, "can't read edited file");
 
-	return rediff (argv[optind], argv[optind + 1], stdout);
+	if (inplace_mode) {
+		/* For in-place mode, write result back to the edited file */
+		FILE *temp_output = xtmpfile();
+		int result = rediff (argv[optind], argv[optind + 1], temp_output);
+
+		if (result == 0) {
+			if (write_file_inplace(argv[optind + 1], temp_output) != 0) {
+				error (EXIT_FAILURE, errno, "failed to write %s", argv[optind + 1]);
+			}
+		}
+
+		fclose (temp_output);
+		return result;
+	} else {
+		return rediff (argv[optind], argv[optind + 1], stdout);
+	}
 }
