@@ -17,7 +17,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  *
  */
 
@@ -140,7 +140,7 @@ void patlist_add_file(struct patlist **dst, const char *fn)
 	char *line = NULL;
 	size_t linelen = 0;
 	size_t len;
-	
+
 	fd = fopen (fn, "r");
 	if (NULL == fd)
 		return;
@@ -158,7 +158,7 @@ void patlist_add_file(struct patlist **dst, const char *fn)
 			line[len - 1] = '\0';
 		}
 		patlist_add (dst, line);
-	} 
+	}
 	fclose (fd);
 }
 
@@ -243,11 +243,11 @@ FILE *xopen_unzip (const char *name, const char *mode)
 	}
 	if (zprog == NULL)
 		return xopen_seekable (name, mode);
-	
+
 	buffer = xmalloc (buflen);
 	fo = xtmpfile();
-	fi = xpipe(zprog, &pid, "r", zprog, name, NULL);
-	
+	fi = xpipe(zprog, &pid, "r", (char **) (const char *[]) { zprog, name, NULL });
+
 	while (!feof (fi)) {
 		size_t count = fread (buffer, 1, buflen, fi);
 		if (ferror (fi)) {
@@ -256,17 +256,17 @@ FILE *xopen_unzip (const char *name, const char *mode)
 		}
 		if (count < 1)
 			break;
-		
+
 		fwrite (buffer, count, 1, fo);
 		if (ferror (fo))
 			error (EXIT_FAILURE, errno, "writing temp file");
 
 		any_data = 1;
 	}
-	
+
 	free (buffer);
 	fclose (fi);
-	
+
 	waitpid (pid, &status, 0);
 	if (any_data == 0 && WEXITSTATUS (status) != 0)
 	{
@@ -280,29 +280,18 @@ FILE *xopen_unzip (const char *name, const char *mode)
 }
 
 /* safe pipe/popen.  mode is either "r" or "w" */
-FILE * xpipe(const char * cmd, pid_t *pid, const char *mode, ...)
+FILE * xpipe(const char * cmd, pid_t *pid, const char *mode, char *const argv[])
 {
-	va_list ap;
 	int fildes[2];
 	int child;
-	int nargs = 0;
-	char *argv[128], *arg;
 	FILE *res;
-	
+
 	if (!mode || (*mode != 'r' && *mode != 'w'))
 		error (EXIT_FAILURE, 0, "xpipe: bad mode: %s", mode);
-	
-	va_start(ap, mode);
-	do {
-		arg = va_arg(ap, char *);
-		argv[nargs++] = arg;
-		if (nargs >= 128)
-			error (EXIT_FAILURE, 0, "xpipe: too many args");
-	} while (arg != NULL);
-	va_end(ap);
-	
+
 	fflush (NULL);
-	pipe (fildes);
+	if (pipe (fildes) == -1)
+		error (EXIT_FAILURE, errno, "pipe failed");
 	child = fork ();
 	if (child == -1) {
 		perror ("fork");
@@ -312,14 +301,17 @@ FILE * xpipe(const char * cmd, pid_t *pid, const char *mode, ...)
 		if (*mode == 'r') {
 			close (fildes[0]);
 			close (1);
-			dup (fildes[1]);
+			if (dup (fildes[1]) == -1)
+				error (EXIT_FAILURE, errno, "dup failed");
 			close (fildes[1]);
 		} else {
 			close (fildes[1]);
 			close (1);
-			dup(2);
+			if (dup(2) == -1)
+				error (EXIT_FAILURE, errno, "dup failed");
 			close (0);
-			dup (fildes[0]);
+			if (dup (fildes[0]) == -1)
+				error (EXIT_FAILURE, errno, "dup failed");
 			close (fildes[0]);
 		}
 		execvp (cmd, argv);
@@ -328,7 +320,7 @@ FILE * xpipe(const char * cmd, pid_t *pid, const char *mode, ...)
 	}
 	if (pid != NULL)
 		*pid = child;
-	
+
 	if (*mode == 'r') {
 		close (fildes[1]);
 		res = fdopen (fildes[0], "r");
@@ -342,49 +334,7 @@ FILE * xpipe(const char * cmd, pid_t *pid, const char *mode, ...)
 	return res;
 }
 
-/*
- * stuff needed for non-GNU systems
- */
 
-#ifndef HAVE_GETLINE
-#define GLSTEP 512
-
-/* suboptimal implementation of glibc's getline() */
-ssize_t getline(char **line, size_t *n, FILE *f)
-{
-	char *p;
-	size_t len;
-	
-	if (*line == NULL || *n < 2) {
-		p = realloc(*line, GLSTEP);
-		if (!p)
-			return -1;
-		*line = p;
-		*n = GLSTEP;
-	}
-	
-	p = fgets(*line, *n, f);
-	if (!p)
-		return -1;
-	
-	len = strlen(p);
-	while ((*line)[len - 1] != '\n') {
-		p = realloc(*line, *n + GLSTEP);
-		if (!p)
-			return -1;
-		*line = p;
-		*n += GLSTEP;
-		
-		p = fgets(p + len, *n - len, f);
-		if (!p)
-			break;
-
-		len = len + strlen(p);
-	}
-	return (ssize_t) len;
-}
-
-#endif
 
 char *progname = "(null)"; /* for error() */
 void set_progname(const char *s)
@@ -392,3 +342,74 @@ void set_progname(const char *s)
 	progname = xstrdup(s);
 }
 
+/* Safe in-place file writing using atomic rename */
+int write_file_inplace(const char *filename, FILE *content)
+{
+	char *temp_name = NULL;
+	FILE *temp_file = NULL;
+	int temp_fd = -1;
+	int ret = -1;
+	size_t filename_len;
+	const char temp_suffix[] = ".tmp.XXXXXX";
+
+	if (!filename || !content) {
+		error(0, 0, "write_file_inplace: invalid arguments");
+		return -1;
+	}
+
+	/* Create temporary filename */
+	filename_len = strlen(filename);
+	temp_name = xmalloc(filename_len + sizeof(temp_suffix));
+	strcpy(temp_name, filename);
+	strcat(temp_name, temp_suffix);
+
+	/* Create temporary file */
+	temp_fd = xmkstemp(temp_name);
+	temp_file = fdopen(temp_fd, "w");
+	if (!temp_file) {
+		error(0, errno, "failed to open temporary file %s", temp_name);
+		close(temp_fd);
+		unlink(temp_name);
+		goto cleanup;
+	}
+
+	/* Copy content to temporary file */
+	rewind(content);
+	while (!feof(content)) {
+		int ch = fgetc(content);
+		if (ch == EOF)
+			break;
+		if (fputc(ch, temp_file) == EOF) {
+			error(0, errno, "failed to write to temporary file %s", temp_name);
+			goto cleanup;
+		}
+	}
+
+	/* Ensure all data is written */
+	if (fflush(temp_file) != 0) {
+		error(0, errno, "failed to flush temporary file %s", temp_name);
+		goto cleanup;
+	}
+
+	fclose(temp_file);
+	temp_file = NULL;
+
+	/* Atomically replace original file */
+	if (rename(temp_name, filename) != 0) {
+		error(0, errno, "failed to rename %s to %s", temp_name, filename);
+		goto cleanup;
+	}
+
+	ret = 0; /* success */
+
+cleanup:
+	if (temp_file)
+		fclose(temp_file);
+	if (temp_name) {
+		if (ret != 0)
+			unlink(temp_name); /* cleanup on failure */
+		free(temp_name);
+	}
+	return ret;
+}
+
