@@ -1,5 +1,8 @@
 [ -n "$VERBOSE" ] && set -x
-rm -rf ${ARENA:=test-arena}
+: ${ARENA:=test-arena}
+THISDIR=$(dirname $0)
+ARENA=${ARENA}/${THISDIR##*/}
+rm -rf $ARENA
 mkdir -p $ARENA
 top_builddir=`cd ${top_builddir-.}; pwd`
 INTERDIFF=${top_builddir}/src/interdiff
@@ -7,6 +10,7 @@ REDIFF=${top_builddir}/src/rediff
 COMBINEDIFF=${top_builddir}/src/combinediff
 FLIPDIFF=${top_builddir}/src/flipdiff
 LSDIFF=${top_builddir}/src/lsdiff
+PATCHVIEW=${top_builddir}/src/patchview
 GREPDIFF=${top_builddir}/src/grepdiff
 FILTERDIFF=${top_builddir}/src/filterdiff
 SELECTDIFF=${top_builddir}/src/selectdiff
@@ -17,5 +21,7 @@ export LSDIFF
 : ${DIFF:=diff}
 : ${PATCH:=patch -s}
 cd $ARENA
-export PS4='[$LINENO] '
+if [ -n "$LxINENO" ]; then
+  export PS4='[$LINENO] '
+fi
 set -x
